//Copyright  2015 Max "Shrinker" Wieden
//Refer to license for license details.
[@vertex330]
layout(std140) uniform matrices
{
	mat4
		modelviewMatrix ,
		projectionMatrix;
};

in vec3
	vPosition,
	vEdges1  ;
in vec2 vEdges2;
in vec3 vNormal;
in int  vFace  ;

out vec3       edges1;
out vec2       edges2;
flat out float light ;
flat out int   face  ;
out vec3       pos   ;

#ifndef WA_NVIDIA
const float
#else
float
#endif
	Pi = 2*asin(1.f);

const float
	SunAmbient =    .8f,
	SunHeading =  30   ,
	SunPitch   = -60   ;

void main()
{
	gl_Position = projectionMatrix*modelviewMatrix*vec4(vPosition, 1);
	pos         = vPosition;
	edges1      = vec3(vEdges1.x == -1 ? 1000.f : vEdges1.x, vEdges1.y == -1 ? 1000.f : vEdges1.y, vEdges1.z == -1 ? 1000.f : vEdges1.z);
	edges2      = vec2(vEdges2.x == -1 ? 1000.f : vEdges2.x, vEdges2.y == -1 ? 1000.f : vEdges2.y);
#ifndef WA_NVIDIA
	const float
#else
	float
#endif
		p = Pi*SunPitch/180,
		t = cos(p)         ;
	light = SunAmbient-dot(vNormal, vec3(-sin(Pi*SunHeading/180)*t, sin(p), -cos(Pi*SunHeading/180)*t))*(1-SunAmbient);
	face  = vFace;
}

[@fragment330]
in      vec3  edges1;
in      vec2  edges2;
flat in float light ;
flat in int   face  ;
in      vec3  pos   ;

out vec4 fragColor;

void main()
{
	float
		val       = min(min(min(edges1.x, edges1.y), edges1.z), min(edges2.x, edges2.y)), //smallest distance to a border
		maxBorder =
			8*min(min(
				length(dFdx(pos)),
				length(dFdy(pos))), 1);
	val       = (val < maxBorder ? (1-pow(1-val/maxBorder, 3))*.75f : .75f)+.25f;
	fragColor =
		vec4(
			face == 0
			? vec3(light*val)
			//selected
			: light*val*
				mix(
					vec3(1), vec3(.9f, .5f, .5f),
					//red fact
					face == 1                                   || //main selection set
					face == 2 && int(gl_FragCoord.x/2) % 2 == 1 || //selection set 1
					face == 3 && int(gl_FragCoord.y/2) % 2 == 1    //selection set 2
					? 1.f
					: .65f), 1);
}
