//Copyright  2015 Max "Shrinker" Wieden
//Refer to license for license details.
[@vertex330]
layout(std140) uniform matrices
{
	mat4
		modelviewMatrix ,
		projectionMatrix;
};

uniform vec3  axisX                  ;
uniform vec3  axisY                  ;
uniform vec3  axisZ                  ;
uniform vec3  camPosition            ;
uniform int   cellsPerMajorStep      ;
uniform int   circleResolution       ;
uniform int   circleResolutionBase   ;
uniform bool  cursorSnapToCenterLines;
uniform bool  cursorSnapToXYU        ;
uniform bool  cursorSnapToXYV        ;
uniform bool  cursorSnapToYZU        ;
uniform bool  cursorSnapToYZV        ;
uniform bool  cursorSnapToZXU        ;
uniform bool  cursorSnapToZXV        ;
uniform int   lookatPlane            ;
uniform int   majorCountPerSideAtRes0;
uniform vec3  origin                 ;
uniform float planarOffset           ;
uniform int   planeSetup             ;
uniform int   resolution             ;
uniform int   resolutionBase         ;
uniform int   shape                  ;

#ifndef WA_ATI
#else
in int _VertexID;
#endif

smooth out vec2  uv         ;
flat   out vec2  uvStart    ;
flat   out int   plane      ;
flat   out float gapFact    ;
flat   out float gapFactC   ;
flat   out int   bULinesFull;
flat   out int   bVLinesFull;

float fact;

bool decomposePoint(const in vec3 u, const in vec3 v, const in vec3 w, const in vec3 p, out vec3 params)
{
	float div = dot(u, cross(v, w));
	if (abs(div) >= 0.00001)
	{
		params = vec3(
			dot(p, cross(v, w))/div,
			dot(p, cross(w, u))/div,
			dot(p, cross(u, v))/div);
		return true;
	}
#ifndef WA_NVIDIA
	params = vec3(0);
	return false;
#else
	else
	{
		params = vec3(0);
		return false;
	}
#endif
}

float planePositionParam(const in float param)
{
	if (param >= 0)
		return max(param, fact/2)+fact/2;
	return min(param, -fact/2)-fact/2;
}

float sign2(const in float x)
{
	if (x >= 0)
		return -1.f;
	return 1.f;
}

void main()
{
#ifndef WA_ATI
	int
		div = gl_VertexID/4,
		mod = gl_VertexID%4;
#else
	int
		div = _VertexID/4,
		mod = _VertexID%4;
#endif
	fact = cellsPerMajorStep*majorCountPerSideAtRes0;
	gapFact = pow(resolutionBase, -resolution);
	if (resolution < 0)
		fact *= gapFact;
	vec3 camUVW;
	if (!decomposePoint(axisX, axisY, axisZ, camPosition-origin, camUVW))
		camUVW = vec3(0);
	vec3 pos = origin;
	uv = vec2(0);
	if (planeSetup == 0)
	{
		plane = div;
		if (plane == 0)
		{
			if (mod == 1 || mod == 2) {float u = planePositionParam(camUVW.z); pos += axisZ*u; uv.x += u;}
			if (mod >= 2            ) {float v = planePositionParam(camUVW.x); pos += axisX*v; uv.y += v;}
			if (camUVW.z >= 0) uvStart.x = max(camUVW.z-fact/2, 0); else uvStart.x = min(camUVW.z+fact/2, 0);
			if (camUVW.x >= 0) uvStart.y = max(camUVW.x-fact/2, 0); else uvStart.y = min(camUVW.x+fact/2, 0);
			if (shape == 0)
			{
				bULinesFull = cursorSnapToZXU ? 1 : 0;
				bVLinesFull = cursorSnapToZXV ? 1 : 0;
			}
		}
		else
		if (plane == 1)
		{
			if (mod == 1 || mod == 2) {float u = planePositionParam(camUVW.x); pos += axisX*u; uv.x += u;}
			if (mod >= 2            ) {float v = planePositionParam(camUVW.y); pos += axisY*v; uv.y += v;}
			if (camUVW.x >= 0) uvStart.x = max(camUVW.x-fact/2, 0); else uvStart.x = min(camUVW.x+fact/2, 0);
			if (camUVW.y >= 0) uvStart.y = max(camUVW.y-fact/2, 0); else uvStart.y = min(camUVW.y+fact/2, 0);
			if (shape == 0)
			{
				bULinesFull = cursorSnapToXYU ? 1 : 0;
				bVLinesFull = cursorSnapToXYV ? 1 : 0;
			}
		}
		else
		{
			if (mod == 1 || mod == 2) {float u = planePositionParam(camUVW.y); pos += axisY*u; uv.x += u;}
			if (mod >= 2            ) {float v = planePositionParam(camUVW.z); pos += axisZ*v; uv.y += v;}
			if (camUVW.y >= 0) uvStart.x = max(camUVW.y-fact/2, 0); else uvStart.x = min(camUVW.y+fact/2, 0);
			if (camUVW.z >= 0) uvStart.y = max(camUVW.z-fact/2, 0); else uvStart.y = min(camUVW.z+fact/2, 0);
			if (shape == 0)
			{
				bULinesFull = cursorSnapToYZU ? 1 : 0;
				bVLinesFull = cursorSnapToYZV ? 1 : 0;
			}
		}
		pos +=
			sign(camUVW.y)*axisY*planarOffset*pow(resolutionBase, -resolution)+
			sign(camUVW.z)*axisZ*planarOffset*pow(resolutionBase, -resolution)+
			sign(camUVW.x)*axisX*planarOffset*pow(resolutionBase, -resolution);
	}
	else
	{
		plane = lookatPlane;
		if (plane == 0)
		{
			if (div == 0)
			{
				if (mod == 1 || mod == 2) {float u = planePositionParam(camUVW.z); pos += axisZ*u; uv.x += u;}
				if (mod >= 2            ) {float v = planePositionParam(camUVW.x); pos += axisX*v; uv.y += v;}
				if (camUVW.z >= 0) uvStart.x = max(camUVW.z-fact/2, 0); else uvStart.x = min(camUVW.z+fact/2, 0);
				if (camUVW.x >= 0) uvStart.y = max(camUVW.x-fact/2, 0); else uvStart.y = min(camUVW.x+fact/2, 0);
			}
			else
			if (div == 1)
			{
				if (mod == 1 || mod == 2) {float u = fact*        sign2(camUVW.z); pos += axisZ*u; uv.x += u;}
				if (mod >= 2            ) {float v = planePositionParam(camUVW.x); pos += axisX*v; uv.y += v;}
				uvStart.x = 0;
				if (camUVW.x >= 0) uvStart.y = max(camUVW.x-fact/2, 0); else uvStart.y = min(camUVW.x+fact/2, 0);
			}
			else
			if (div == 2)
			{
				if (mod == 1 || mod == 2) {float u = planePositionParam(camUVW.z); pos += axisZ*u; uv.x += u;}
				if (mod >= 2            ) {float v = fact*        sign2(camUVW.x); pos += axisX*v; uv.y += v;}
				if (camUVW.z >= 0) uvStart.x = max(camUVW.z-fact/2, 0); else uvStart.x = min(camUVW.z+fact/2, 0);
				uvStart.y = 0;
			}
			else //div == 3
			{
				if (mod == 1 || mod == 2) {float u = sign2(camUVW.z)*fact; pos += axisZ*u; uv.x += u;}
				if (mod >= 2            ) {float v = sign2(camUVW.x)*fact; pos += axisX*v; uv.y += v;}
				uvStart.x = 0;
				uvStart.y = 0;
			}
			pos += sign(camUVW.y)*axisY*planarOffset*pow(resolutionBase, -resolution);
			if (shape == 0)
			{
				bULinesFull = cursorSnapToZXU ? 1 : 0;
				bVLinesFull = cursorSnapToZXV ? 1 : 0;
			}
		}
		else
		if (plane == 1)
		{
			if (div == 0)
			{
				if (mod == 1 || mod == 2) {float u = planePositionParam(camUVW.x); pos += axisX*u; uv.x += u;}
				if (mod >= 2            ) {float v = planePositionParam(camUVW.y); pos += axisY*v; uv.y += v;}
				if (camUVW.x >= 0) uvStart.x = max(camUVW.x-fact/2, 0); else uvStart.x = min(camUVW.x+fact/2, 0);
				if (camUVW.y >= 0) uvStart.y = max(camUVW.y-fact/2, 0); else uvStart.y = min(camUVW.y+fact/2, 0);
			}
			else
			if (div == 1)
			{
				if (mod == 1 || mod == 2) {float u = fact*        sign2(camUVW.x); pos += axisX*u; uv.x += u;}
				if (mod >= 2            ) {float v = planePositionParam(camUVW.y); pos += axisY*v; uv.y += v;}
				uvStart.x = 0;
				if (camUVW.y >= 0) uvStart.y = max(camUVW.y-fact/2, 0); else uvStart.y = min(camUVW.y+fact/2, 0);
			}
			else
			if (div == 2)
			{
				if (mod == 1 || mod == 2) {float u = planePositionParam(camUVW.x); pos += axisX*u; uv.x += u;}
				if (mod >= 2            ) {float v = fact*        sign2(camUVW.y); pos += axisY*v; uv.y += v;}
				if (camUVW.x >= 0) uvStart.x = max(camUVW.x-fact/2, 0); else uvStart.x = min(camUVW.x+fact/2, 0);
				uvStart.y = 0;
			}
			else //div == 3
			{
				if (mod == 1 || mod == 2) {float u = sign2(camUVW.x)*fact; pos += axisX*u; uv.x += u;}
				if (mod >= 2            ) {float v = sign2(camUVW.y)*fact; pos += axisY*v; uv.y += v;}
				uvStart.x = 0;
				uvStart.y = 0;
			}
			pos += sign(camUVW.z)*axisZ*planarOffset*pow(resolutionBase, -resolution);
			if (shape == 0)
			{
				bULinesFull = cursorSnapToXYU ? 1 : 0;
				bVLinesFull = cursorSnapToXYV ? 1 : 0;
			}
		}
		else //plane == 2
		{
			if (div == 0)
			{
				if (mod == 1 || mod == 2) {float u = planePositionParam(camUVW.y); pos += axisY*u; uv.x += u;}
				if (mod >= 2            ) {float v = planePositionParam(camUVW.z); pos += axisZ*v; uv.y += v;}
				if (camUVW.y >= 0) uvStart.x = max(camUVW.y-fact/2, 0); else uvStart.x = min(camUVW.y+fact/2, 0);
				if (camUVW.z >= 0) uvStart.y = max(camUVW.z-fact/2, 0); else uvStart.y = min(camUVW.z+fact/2, 0);
			}
			else
			if (div == 1)
			{
				if (mod == 1 || mod == 2) {float u = fact*        sign2(camUVW.y); pos += axisY*u; uv.x += u;}
				if (mod >= 2            ) {float v = planePositionParam(camUVW.z); pos += axisZ*v; uv.y += v;}
				uvStart.x = 0;
				if (camUVW.z >= 0) uvStart.y = max(camUVW.z-fact/2, 0); else uvStart.y = min(camUVW.z+fact/2, 0);
			}
			else
			if (div == 2)
			{
				if (mod == 1 || mod == 2) {float u = planePositionParam(camUVW.y); pos += axisY*u; uv.x += u;}
				if (mod >= 2            ) {float v = fact*        sign2(camUVW.z); pos += axisZ*v; uv.y += v;}
				if (camUVW.y >= 0) uvStart.x = max(camUVW.y-fact/2, 0); else uvStart.x = min(camUVW.y+fact/2, 0);
				uvStart.y = 0;
			}
			else //div == 3
			{
				if (mod == 1 || mod == 2) {float u = sign2(camUVW.y)*fact; pos += axisY*u; uv.x += u;}
				if (mod >= 2            ) {float v = sign2(camUVW.z)*fact; pos += axisZ*v; uv.y += v;}
				uvStart.x = 0;
				uvStart.y = 0;
			}
			pos += sign(camUVW.x)*axisX*planarOffset*pow(resolutionBase, -resolution);
			if (shape == 0)
			{
				bULinesFull = cursorSnapToYZU ? 1 : 0;
				bVLinesFull = cursorSnapToYZV ? 1 : 0;
			}
		}
	}
	if (shape == 1)
	{
		gapFactC = pow(circleResolutionBase, -circleResolution);
		bULinesFull = cursorSnapToCenterLines ? 1 : 0;
		bVLinesFull = bULinesFull;
	}
	gl_Position = projectionMatrix*modelviewMatrix*vec4(pos, 1);
}

[@fragment330]
uniform float axisLineWidth          ;
uniform vec3  axisXNColor            ;
uniform vec3  axisXPColor            ;
uniform vec3  axisYNColor            ;
uniform vec3  axisYPColor            ;
uniform vec3  axisZNColor            ;
uniform vec3  axisZPColor            ;
uniform int   cellsPerMajorStep      ;
uniform int   cellsPerMajorStepAround;
uniform bool  cursorSnapToCenterLines;
uniform bool  cursorSnapToCircles    ;
uniform vec3  majorColor             ;
uniform int   majorCountAroundAtRes0 ;
uniform float majorLineWidth         ;
uniform vec3  minorColor             ;
uniform float minorLineWidth         ;
uniform int   shape                  ;

smooth in vec2  uv         ;
flat   in vec2  uvStart    ;
flat   in int   plane      ;
flat   in float gapFact    ;
flat   in float gapFactC   ;
flat   in int   bULinesFull;
flat   in int   bVLinesFull;

out vec4 fragColor;

#ifndef WA_NVIDIA
const float
#else
float
#endif
	Pi = 2*asin(1.f);

vec2
	uvDeltaX, //direction of line in uv space whose projection on screen (scanline) is perfectly horizontal
	uvDeltaY; //... vertical

bool intersection(const in vec2 l1_p0, const in vec2 l1_dir, const in vec2 l2_p0, const in vec2 l2_dir, out vec2 result)
{
	float div = l1_dir.y*l2_dir.x-l2_dir.y*l1_dir.x;
	if (abs(div) >= 0.00001f)
	{
		result = l1_p0+l1_dir*((l2_p0.y*l2_dir.x+l2_dir.y*(l1_p0.x-l2_p0.x)-l1_p0.y*l2_dir.x)/div);
		return true;
	}
	else
	{
		result = vec2(0);
		return false;
	}
}

bool intersectionU(const in float l1_p0y, const in vec2 l2_p0, const in vec2 l2_dir, out vec2 result)
{
	if (abs(l2_dir.y) >= 0.00001f)
	{
		result = vec2(0, l1_p0y)-vec2((l2_p0.y*l2_dir.x+l2_dir.y*(-l2_p0.x)-l1_p0y*l2_dir.x)/l2_dir.y, 0);
		return true;
	}
	else
	{
		result = vec2(0);
		return false;
	}
}

bool intersectionV(const in float l1_p0x, const in vec2 l2_p0, const in vec2 l2_dir, out vec2 result)
{
	if (abs(l2_dir.x) >= 0.00001f)
	{
		result = vec2(l1_p0x, 0)+vec2(0, (l2_p0.y*l2_dir.x+l2_dir.y*(l1_p0x-l2_p0.x))/l2_dir.x);
		return true;
	}
	else
	{
		result = vec2(0);
		return false;
	}
}

//lineOrigin in world units, nDirection normalized, lineWidth in pixels, gap in world units
bool parallelLines(const in vec2 origin, const in vec2 nDirection, in float lineWidth, const in vec2 gap)
{
	lineWidth /= 2;

	//horizontal (x) scanline
	{
		//determine intersection of scanline (perfectly horizontal in screen space) in uv space with one of the target lines
		vec2 s;
		if (intersection(origin, nDirection, uv, uvDeltaX, s)) //intersection with one of the target lines in uv space
		{
			vec2 perpLine = s-uv;
			//shift the intersection point to be on the closest target line
			if (gap != vec2(0))
			{
				float lineLen = length(perpLine);
				//gap between two lines = distance of intersection points with two neighboring lines, where one line was offset with gap
				vec2 s2;
				intersection(origin+gap, nDirection, uv, uvDeltaX, s2);
				float gapInLine = distance(s, s2);
				//new line length*normalized line
				s = uv+(lineLen-floor(lineLen/gapInLine+.5f)*gapInLine)*normalize(perpLine);
			}
			if (distance(s, uv) <= lineWidth*length(uvDeltaX))
				return true;
		}
	}

	//vertical (y) scanline
	{
		//determine intersection of scanline (perfectly vertical in screen space) in uv space with one of the target lines
		vec2 s;
		if (intersection(origin, nDirection, uv, uvDeltaY, s)) //intersection with one of the target lines in uv space
		{
			vec2 perpLine = s-uv;
			//shift the intersection point to be on the closest target line
			if (gap != vec2(0))
			{
				float lineLen = length(perpLine);
				//gap between two lines = distance of intersection points with two neighboring lines, where one line was offset with gap
				vec2 s2;
				intersection(origin+gap, nDirection, uv, uvDeltaY, s2);
				float gapInLine = distance(s, s2);
				//new line length*normalized line
				s = uv+(lineLen-floor(lineLen/gapInLine+.5f)*gapInLine)*normalize(perpLine);
			}
			if (distance(s, uv) <= lineWidth*length(uvDeltaY))
				return true;
		}
	}

	return false;
}

//lineOrigin in world units, nDirection normalized, lineWidth in pixels, gap in world units
bool parallelLinesU(in float lineWidth, const in float gap)
{
	lineWidth /= 2;

	//horizontal (x) scanline
	{
		//determine intersection of scanline (perfectly horizontal in screen space) in uv space with one of the target lines
		vec2 s;
		if (intersectionU(0.f, uv, uvDeltaX, s)) //intersection with one of the target lines in uv space
		{
			vec2 perpLine = s-uv;
			//shift the intersection point to be on the closest target line
			if (gap != 0)
			{
				float lineLen = length(perpLine);
				//gap between two lines = distance of intersection points with two neighboring lines, where one line was offset with gap
				vec2 s2;
				intersectionU(gap, uv, uvDeltaX, s2);
				float gapInLine = distance(s, s2);
				//new line length*normalized line
				s = uv+(lineLen-floor(lineLen/gapInLine+.5f)*gapInLine)*normalize(perpLine);
			}
			if (distance(s, uv) <= lineWidth*length(uvDeltaX))
				return true;
		}
	}

	//vertical (y) scanline
	{
		//determine intersection of scanline (perfectly vertical in screen space) in uv space with one of the target lines
		vec2 s;
		if (intersectionU(0.f, uv, uvDeltaY, s)) //intersection with one of the target lines in uv space
		{
			vec2 perpLine = s-uv;
			//shift the intersection point to be on the closest target line
			if (gap != 0)
			{
				float lineLen = length(perpLine);
				//gap between two lines = distance of intersection points with two neighboring lines, where one line was offset with gap
				vec2 s2;
				intersectionU(gap, uv, uvDeltaY, s2);
				float gapInLine = distance(s, s2);
				//new line length*normalized line
				s = uv+(lineLen-floor(lineLen/gapInLine+.5f)*gapInLine)*normalize(perpLine);
			}
			if (distance(s, uv) <= lineWidth*length(uvDeltaY))
				return true;
		}
	}

	return false;
}

//lineOrigin in world units, nDirection normalized, lineWidth in pixels, gap in world units
bool parallelLinesV(in float lineWidth, const in float gap)
{
	lineWidth /= 2;

	bool result = false;

	//horizontal (x) scanline
	{
		//determine intersection of scanline (perfectly horizontal in screen space) in uv space with one of the target lines
		vec2 s;
		if (intersectionV(0.f, uv, uvDeltaX, s)) //intersection with one of the target lines in uv space
		{
			vec2 perpLine = s-uv;
			//shift the intersection point to be on the closest target line
			if (gap != 0)
			{
				float lineLen = length(perpLine);
				//gap between two lines = distance of intersection points with two neighboring lines, where one line was offset with gap
				vec2 s2;
				intersectionV(gap, uv, uvDeltaX, s2);
				float gapInLine = distance(s, s2);
				//new line length*normalized line
				s = uv+(lineLen-floor(lineLen/gapInLine+.5f)*gapInLine)*normalize(perpLine);
			}
			if (distance(s, uv) <= lineWidth*length(uvDeltaX))
				return true;
		}
	}

	//vertical (y) scanline
	{
		//determine intersection of scanline (perfectly vertical in screen space) in uv space with one of the target lines
		vec2 s;
		if (intersectionV(0.f, uv, uvDeltaY, s)) //intersection with one of the target lines in uv space
		{
			vec2 perpLine = s-uv;
			//shift the intersection point to be on the closest target line
			if (gap != 0)
			{
				float lineLen = length(perpLine);
				//gap between two lines = distance of intersection points with two neighboring lines, where one line was offset with gap
				vec2 s2;
				intersectionV(gap, uv, uvDeltaY, s2);
				float gapInLine = distance(s, s2);
				//new line length*normalized line
				s = uv+(lineLen-floor(lineLen/gapInLine+.5f)*gapInLine)*normalize(perpLine);
			}
			if (distance(s, uv) <= lineWidth*length(uvDeltaY))
				return true;
		}
	}

	return false;
}

//x in world units
bool stripeCheck(in float x, in float fact)
{
	float wPerc = .25f+pow(gl_FragCoord.z, 32*fact)*.8f;
	return fract(2*x/fact+wPerc/2) < wPerc;
}

//origin in world units, lineWidth in pixels, gap in world units
bool circles(const in vec2 origin, const in float lineWidth, const in float gap)
{
	if (!cursorSnapToCircles)
	{
		vec2 d = uv-origin;
#ifndef WA_ATI
		float a = atan(d.y, d.x);
#else
		float a;
		if
		(
			d.y >= 0 && ( d.y <= d.x ||  d.y <= -d.x) ||
			d.y <= 0 && (-d.y <= d.x || -d.y <= -d.x)
		)
			a = atan(d.x, -d.y)-Pi/2;
		else
			a = atan(d.y, d.x);
#endif
		if (!stripeCheck(a/(2*Pi)*(cellsPerMajorStepAround*majorCountAroundAtRes0), gapFactC))
			return false;
	}
	vec2 dn = normalize(uv-origin);
	return parallelLines(origin, vec2(dn.y, -dn.x), lineWidth, dn*gap);
}

//origin in world units, lineWidth in pixels
bool centerLines(const in vec2 origin, const in float lineWidth, const in int count)
{
	if (cursorSnapToCenterLines || stripeCheck(distance(origin, uv), gapFact))
	{
		vec2 d = uv-origin;
#ifndef WA_ATI
		float a = 2*Pi*floor((atan(d.y, d.x)*count/Pi+1)/2)/count;
#else
		float a;
		if
		(
			d.y >= 0 && ( d.y <= d.x ||  d.y <= -d.x) ||
			d.y <= 0 && (-d.y <= d.x || -d.y <= -d.x)
		)
			a = atan(d.x, -d.y)-Pi/2;
		else
			a = atan(d.y, d.x);
		a = 2*Pi*floor((a*count/Pi+1)/2)/count;
#endif
		return parallelLines(origin, vec2(cos(a), sin(a)), lineWidth, vec2(0));
	}
	return false;
}

void main()
{
	uvDeltaX = dFdx(uv);
	uvDeltaY = dFdy(uv);

	bool
		uInRange = uvStart.x >= 0 && uv.x >= uvStart.x || uvStart.x <= 0 && uv.x <= uvStart.x,
		vInRange = uvStart.y >= 0 && uv.y >= uvStart.y || uvStart.y <= 0 && uv.y <= uvStart.y;
#ifndef WA_ATI
#else
	bool set = false;
#endif
	//
	// axes
	//
	if (plane == 0)
	{
		if (abs(uv.x) >= abs(uv.y))
		{
			if
			(
				uInRange &&
				(bULinesFull != 0 || stripeCheck(uv.x, gapFact)) &&
				parallelLinesU(axisLineWidth, 0.f)
			)
			{
				if (uv.x >= 0)
					fragColor = vec4(axisZPColor, 1);
				else
					fragColor = vec4(axisZNColor, 1);
#ifndef WA_ATI
				return;
#else
				set = true;
#endif
			}
		}
		else
			if
			(
				vInRange &&
				(bVLinesFull != 0 || stripeCheck(uv.y, gapFact)) &&
				parallelLinesV(axisLineWidth, 0.f)
			)
			{
				if (uv.y >= 0)
					fragColor = vec4(axisXPColor, 1);
				else
					fragColor = vec4(axisXNColor, 1);
#ifndef WA_ATI
				return;
#else
				set = true;
#endif
			}
	}
	else
	if (plane == 1)
	{
		if (abs(uv.x) >= abs(uv.y))
		{
			if
			(
				uInRange &&
				(bULinesFull != 0 || stripeCheck(uv.x, gapFact)) &&
				parallelLinesU(axisLineWidth, 0.f)
			)
			{
				if (uv.x >= 0)
					fragColor = vec4(axisXPColor, 1);
				else
					fragColor = vec4(axisXNColor, 1);
#ifndef WA_ATI
				return;
#else
				set = true;
#endif
			}
		}
		else
			if
			(
				vInRange &&
				(bVLinesFull != 0 || stripeCheck(uv.y, gapFact)) &&
				parallelLinesV(axisLineWidth, 0.f)
			)
			{
				if (uv.y >= 0)
					fragColor = vec4(axisYPColor, 1);
				else
					fragColor = vec4(axisYNColor, 1);
#ifndef WA_ATI
				return;
#else
				set = true;
#endif
			}
	}
	else //plane == 2
		if (abs(uv.x) >= abs(uv.y))
		{
			if
			(
				uInRange &&
				(bULinesFull != 0 || stripeCheck(uv.x, gapFact)) &&
				parallelLinesU(axisLineWidth, 0.f)
			)
			{
				if (uv.x >= 0)
					fragColor = vec4(axisYPColor, 1);
				else
					fragColor = vec4(axisYNColor, 1);
#ifndef WA_ATI
				return;
#else
				set = true;
#endif
			}
		}
		else
			if
			(
				vInRange &&
				(bVLinesFull != 0 || stripeCheck(uv.y, gapFact)) &&
				parallelLinesV(axisLineWidth, 0.f)
			)
			{
				if (uv.y >= 0)
					fragColor = vec4(axisZPColor, 1);
				else
					fragColor = vec4(axisZNColor, 1);
#ifndef WA_ATI
				return;
#else
				set = true;
#endif
			}
#ifndef WA_ATI
#else
	if (set)
		return;
#endif
	//
	// grid lines
	//
	if (uInRange && vInRange)
		if (shape == 0)
		{
			//majors
			if
			(
				(bULinesFull != 0 || stripeCheck(uv.x, gapFact)) && parallelLinesU(majorLineWidth, gapFact*cellsPerMajorStep) ||
				(bVLinesFull != 0 || stripeCheck(uv.y, gapFact)) && parallelLinesV(majorLineWidth, gapFact*cellsPerMajorStep)
			)
			{
				fragColor = vec4(majorColor, 1);
#ifndef WA_ATI
				return;
#else
				set = true;
#endif
			}
			//minors
			if
			(
				(bULinesFull != 0 || stripeCheck(uv.x, gapFact)) && parallelLinesU(minorLineWidth, gapFact) ||
				(bVLinesFull != 0 || stripeCheck(uv.y, gapFact)) && parallelLinesV(minorLineWidth, gapFact)
			)
			{
#ifndef WA_ATI
#else
				if (!set)
#endif
				fragColor = vec4(minorColor, 1);
#ifndef WA_ATI
				return;
#else
				set = true;
#endif
			}
		}
		else //shape == 1
		{
			//majors
			if
			(
				circles    (vec2(0), majorLineWidth, gapFact*cellsPerMajorStep            ) ||
				centerLines(vec2(0), majorLineWidth, int(majorCountAroundAtRes0/gapFactC))
			)
			{
				fragColor = vec4(majorColor, 1);
#ifndef WA_ATI
				return;
#else
				set = true;
#endif
			}
			//minors
			if
			(
				circles    (vec2(0), minorLineWidth, gapFact                                                      ) ||
				centerLines(vec2(0), minorLineWidth, int(cellsPerMajorStepAround*majorCountAroundAtRes0/gapFactC))
			)
			{
#ifndef WA_ATI
#else
				if (!set)
#endif
				fragColor = vec4(minorColor, 1);
#ifndef WA_ATI
				return;
#else
				set = true;
#endif
			}
		}
#ifndef WA_ATI
#else
	if (set)
		return;
#endif
	discard;
}
